@@ -1,5 +1,5 @@
 /*
- * $Id: VariableProvider.java,v 1.1.1.1 1998/05/11 18:43:03 metlov Exp $
+ * $Id: VariableProvider.java,v 1.3 1998/10/28 23:10:56 metlov Exp $
  *
  * This file is part of the Java Expressions Library (JEL).
  *   For more information about JEL visit :
@@ -24,4 +24,21 @@ public class VariableProvider {
     return 0;
   };
 
+  public static int bool2int(boolean b) {
+    if (b) return 1; else return 0;
+  };
+
+  public static float NaNf() {
+    return Float.NaN;
+  };
+
+  public static double NaNd() {
+    return Double.NaN;
+  };
+
+  public static Double makeDoubleObject(double d) {
+    return new Double(d);
+  };
+
 };
+
