@@ -1,5 +1,5 @@
 /*
- * $Id: Optimizer.java,v 1.2 1998/05/23 12:34:50 metlov Exp $
+ * $Id: Optimizer.java,v 1.6 1998/11/07 19:30:29 metlov Exp $
  *
  * This file is part of the Java Expressions Library (JEL).
  *   For more information about JEL visit :
@@ -20,45 +20,45 @@ import gnu.jel.debug.Tester;
 import java.lang.reflect.Method;
 
 /**
- * This class handles storage of the expressions, user level type 
- * checking and optimizations.
- * <P> The only optimization currently supported is the evaluation of the 
- * constant subexpressions. This class goes even further in its attempts to
- * optimize, it evaluates not only arifmetic constant subexpressions, but 
- * also calls to the static methods of the constant arguments. Namely, 
- * expressions such as sin(1) will be evaluated at compile time.
- * <P> Another function of this class is the error handling. If during 
- * evaluation of the constant subexpression a error occurs (Throwable is 
- * thrown) it is caught and compiled into the resulting class image to be 
- * rethrown at execution time. Example :
- * <TT>1/0</TT> will not cause any exceptions to come out of the compiler, but 
- * will be compiled into <TT>{ throw caughtArithmeticError; }</TT> if the 
- * optimization is turned on. When optimization is off the previous
- * expression will be compiled to <TT>{ return 1/0 }</TT>, which will throw
- * exception at run time ( execution of the <TT>evaluate(...)</TT> method
- * of the <TT>gnu.jel.CompiledExpression</TT>). Another thing to note, that 
- * <TT>Throwable</TT>s caught at compile time are not recreated at compile 
- * time  ( they are not compiled into <TT>{throw <B>new</B> 
- * ArithmeticError()}</TT>). Errors being thrown from <TT>evaluate(..)</TT>
- * method are EXACTLY the same object as those caught during the constant 
- * subexpression eveluation. This approach should greatly simplify the error
- * handling and applies not only to arithmetic expressions but also to any 
- * exceptions, thrown by interpretable methods.
- * <P> For more information on how to specify static non interpretable 
- * method (as in the case of <TT>java.Math.random()</TT> see the 
- * documentation for <TT>gnu.jel.Library</TT>
+ * This class handles storage of the expressions automatic widening type 
+ * conversions and optimizations.
+ * <P> Currently the only supported type of optimization is evaluation of
+ * constant subexpressions. Unlike other compilers JEL not only evaluates
+ * arithmetic operators on constants but also attempts to call some (those 
+ * which marked stateless) functions if all their arguments are constants.
+ * For example, expression "sin(1)" will be completely evaluated at compile
+ * time.
+ * <P> The gnu.jel.Library class handles details on which functions can be 
+ * evaluated at compile time and which can not, see it's documentation for
+ * details.
  * @see gnu.jel.Library
  * @author Konstantin L. Metlov (metlov@fzu.cz)
  */
 public class Optimizer {
+
+  static Class string_class=null;
+  
+  static {
+    try {
+      string_class=Class.forName("java.lang.String");
+    } catch (ClassNotFoundException e2) {
+      if (Debug.enabled)
+	Debug.println("Get Yourself a real JAVA.");
+      // Can't be in java
+    };
+  };
   
   private Library lib;
 
   private Stack types=new Stack();
-  private OPlist code=new OPlist();
+
+  /**
+   * Double linked list of operations in this expression.
+   */
+  protected OPlist code=new OPlist();
   private Stack functions=new Stack();
   private Stack functionsDescriptors=new Stack();  
-  private Throwable caughtThrowable=null;
+  //  private Throwable caughtThrowable=null;
   
   private boolean finished=false;
 
@@ -138,7 +138,14 @@ public class Optimizer {
   public void load(double c) {
     load(Double.TYPE,new Double(c));
   };
-  
+
+  /**
+   * Generates a "load String constant" operation.
+   * @param s is the String constant to load.
+   */
+  public void load(String s) {
+    load(string_class,s);
+  };
   
   private void load(Class t, Object o) {
     seeNonFinished();
@@ -148,7 +155,7 @@ public class Optimizer {
   };
   
   /**
-   * Generates an explicit "convert type" operation.
+   * Generates an explicit type conversion operation.
    * <P> It has the sense to call this function only for narrowing 
    * conversions (see &sect;5.1.3 and &sect;5.1.5 of the Java Language
    * Specification (JLS) ). 
@@ -184,7 +191,7 @@ public class Optimizer {
   };
 
   /**
-   * Generates an explicit "convert type" operation.
+   * Generates an explicit type conversion operation.
    * <P> This function is equivalent to <TT>convert(to,false)</TT>.
    * Widening ( &sect;5.1.2,  &sect;5.1.4 of JLS) type conversions are 
    * performed by this compiler automatically.
@@ -216,6 +223,41 @@ public class Optimizer {
     };
     code.addLast(new OP_unary(o));
   };
+
+  /**
+   * Denotes start of group of logical operators whose result should be inverted.
+   * @see gnu.jel.Optimizer#logical_not
+   */
+  public void logical_not_start() {
+    seeNonFinished();
+    OP_logical_not opc=new OP_logical_not();
+    functions.push(opc);
+    int[] descrs={0};
+    functionsDescriptors.push(descrs);
+    code.addLast(new OP_start(opc));
+  };
+  
+  /**
+   * Inverts result of group of logical operators.
+   * <P> To generate logical not operation it is needed :
+   * <PRE>
+   * 1. logical_not_start
+   * 2. calculate value to be inverted
+   * 3. logical_not
+   * </PRE>
+   * @see gnu.jel.Optimizer#logical_not
+   */
+  public void logical_not() throws IllegalStateException {
+    if (types.peek()!=Boolean.TYPE)
+      throw new IllegalStateException("You tried to use logical complement"+
+				      " on "+((Class)types.peek()).getName()+
+				      ". This operation is supported only"+
+				      " on booleans.");
+    OP_logical_not opf=(OP_logical_not) functions.pop();
+    int[] descrs=(int[])functionsDescriptors.pop();
+    code.addLast(opf);
+  };
+
   
   /**
    * Denotes the start of the function call.
@@ -333,87 +375,200 @@ public class Optimizer {
    * for the types in stack.
    * @see gnu.jel.ExpressionImage
    */
-  public void binaryOP(int o) throws IllegalStateException {
+  public void binaryOP(int o,boolean logical) throws IllegalStateException {
     seeNonFinished();
     OP_binary opf=(OP_binary) functions.peek();
     int[] descrs=(int[])functionsDescriptors.peek();
     
-    code.addLast(new OP_param(opf,(Class)types.peek()));
+    OP_param terminating_param=new OP_param(opf,(Class)types.peek());
+    terminating_param.setbinaryparam=false;
+    code.addLast(terminating_param);
     
     Class op2=(Class)types.pop();
     Class op1=(Class)types.pop();
-    
-    if (!ExpressionImage.canGenerateBinary(o,op1,op2)) {
-      types.push(op1);
-      types.push(op2);
-      throw new IllegalStateException("Types "+op1.toString()+" and "+
-				      op2.toString()+" are not suitable "+
-				      "for the operation \""+
-				      ExpressionImage.binaryNames[o]+"\"");
+
+    // Catch possible errors
+    String message=null;
+    boolean error=false;
+
+    if (!logical) {
+      if (!ExpressionImage.canGenerateBinary(o,op1,op2)) {
+	error=true;
+	message="Types "+op1.toString()+" and "+ op2.toString()+
+	  " are not suitable "+"for the operation \""+
+	  ExpressionImage.binaryNames[o]+"\"";
+      };
+    } else {
+      if ((op1!=Boolean.TYPE) || (op2!=Boolean.TYPE)) {
+	error=true;
+	message="Types "+op1.toString()+" and "+ op2.toString()+
+	  " are not suitable for \""+ExpressionImage.logicalNames[o]+
+	  "\" operation. Both operands must be boolean.";
+      };
     };
     
-    Class ft=ExpressionImage.getMostGeneralType(op1,op2);
+    if (error) {
+      types.push(op1);
+      types.push(op2);
+      throw new IllegalStateException(message);
+    };
+    // all errors reported
     
+    // Determine types of operands and result
+    Class opType=Boolean.TYPE,resType=Boolean.TYPE;
+    if (!logical) {
+      if (ExpressionImage.isPromotionBinary(o)) {
+	if ((o==ExpressionImage.BI_PL) && (op1==string_class))
+	  opType=string_class;
+	else
+	  opType=ExpressionImage.getBinaryPromoted(op1,op2);
+      } else {
+	opType=resType=ExpressionImage.getUnaryPromoted(op1);
+      };
+      resType=opType;
+      if ((o>=ExpressionImage.BI_EQ) && (o<=ExpressionImage.BI_LE)) 
+	resType=Boolean.TYPE;
+    };
+    
+    // Note operation
     OP_binary opc=(OP_binary)opf;    
-    
+      
     code.addLast(opc);
     
-    opc.setOperation(code,o,ft);
-    
-    types.push(ft);
+    opc.setOperation(code,o,logical,opType);
+      
+    types.push(resType);
     functions.pop(); functionsDescriptors.pop();
   };
+
+  /**
+   * Starts generation of conditional ?: .
+   * <P> The stack should contain boolean value.
+   * <P> To generate complete conditional it is needed :
+   * <PRE>
+   * 1. calculate condition (boolean)
+   * 2. conditional_true()
+   * 3. add instructions to executed if condition is "true"
+   * 4. conditional_false()
+   * 5. add instructions to executed if condition is "false"
+   * 6. conditional_end()
+   * </PRE>
+   */
+  public void conditional_true()  throws IllegalStateException {
+    if (types.pop()!=Boolean.TYPE)
+      throw new IllegalStateException("The first operand of conditional must"+
+				      " be of boolean type.");
+
+    OP_conditional opc=new OP_conditional();
+    functions.push(opc);
+    int[] descrs={0};
+    functionsDescriptors.push(descrs);
+    code.addLast(new OP_start(opc));
+  };
+
+  /**
+   * Continues generation of conditional ?: .
+   * @see gnu.jel.Optimizer#conditional_true
+   */
+  public void conditional_false() {
+    seeNonFinished();
+    OP_conditional opf=(OP_conditional) functions.peek();
+    int[] descrs=(int[]) functionsDescriptors.peek();
+    descrs[0]++;
+    
+    // Search for corresponding OP_start
+    OP start=code.getLast();
+    while (!((start instanceof OP_start) && 
+	     (((OP_start)start).getFunction()==opf))) start=start.prev;
+    
+    // create list of instructions of true branch
+    opf.setTrueList(code.cut_end(start.next));
+  };
+
+  /**
+   * Finishes generation of conditional ?: .
+   * @see gnu.jel.Optimizer#conditional_true
+   */
+  public void conditional_end()  throws IllegalStateException  {
+    seeNonFinished();
+    OP_conditional opf=(OP_conditional) functions.peek();
+    int[] descrs=(int[]) functionsDescriptors.peek();
+    descrs[0]++;
+
+    // Search for corresponding OP_start
+    OP start=code.getLast();
+    while (!((start instanceof OP_start) && 
+	     (((OP_start)start).getFunction()==opf))) start=start.prev;
+
+    // create list of instructions of false branch
+    opf.setFalseList(code.cut_end(start.next));
+    
+    // Now what remains is the start instruction. Take it out.
+    code.remove(start);
+    
+    // Let's determine the type of result.
+    // Now We have in stack results of both branches.
+    
+    Class op2=(Class)types.pop();
+    Class op1=(Class)types.pop();
+
+    // According to JLS 15.24
+    Class resType=null;
+    if (op1.isPrimitive() && op2.isPrimitive())
+      resType=ExpressionImage.getBinaryPromoted(op1,op2);
+    else {
+      if (op1==op2) resType=op1;
+      else if (ExpressionImage.canConvertByWidening(op1,op2)) resType=op2;
+      else if (ExpressionImage.canConvertByWidening(op2,op1)) resType=op1;
+    };
+    
+    if (resType==null)
+      throw new IllegalStateException("Operands of conditional have types "+
+				      op1.getName()+" and "+op2.getName()+
+				      " . These types are not compatible.");
+    opf.setType(resType);
+    types.push(resType);
+
+    code.addLast(opf);
+    functions.pop(); functionsDescriptors.pop();    
+  };
+
   
   /**
    * Finishes the function.
    * <P> This method should be called exactly once before the attempt
-   * to compile is made.
-   * <P> If there was nothing in the stack before it's called this method
-   * generates the returning of the null pointer.
-   * @exception  IllegalStateException if the function is not in the
+   * to instantiate expression (or obtain its image) is made.
    * proper state to be finished (some extra items are in stack).
    */
-  public void finish() throws IllegalStateException {
+  public void finish() {
     seeNonFinished();
-    if ((code.size!=0) && (types.size()!=1))
-      throw new IllegalStateException("Stack should contain single item on"+
-				      " exit.");
-    if ((functions.size() !=0) || (functionsDescriptors.size() !=0)) 
-      throw new IllegalStateException("There are function calls in progress.");
-    
+    if (Debug.enabled) {
+      Debug.assert((code.size==0) || (types.size()==1),
+		   "Stack should contain single item on exit.");
+      Debug.assert((functions.size() ==0) && (functionsDescriptors.size() ==0),
+		   "There are function calls in progress.");
+    };
     finished=true;
   };
   
   /**
    * Optimizes the function.
    * <P> Currently only evaluation of constant subexpressions is performed.
-   * @param of is the optimization flags (0 - do nothing; >0 optimize).
+   * @param of is the maximum number of optimization iterations to perform.
    */
   public void optimize(int of) {
     seeFinished();
-    if (of==0) return;
-    while(optimizeIteration());
+    while ((of>0) && optimizeIteration(code)) of--;
   };
   
-  protected boolean optimizeIteration() {
-    OP cop=code.getFirst();
-    boolean didsome=false;
-    while ((cop!=null) && (caughtThrowable==null)) {
-      OP ncop=cop.next();
-      if (cop instanceof OP_function) {
-	OP_function copf=(OP_function) cop;
-	if (copf.canInterpret()) {
-	  try {
-	    copf.interpret(code);
-	    didsome=true;
-	  } catch (Throwable exc) {
-	    caughtThrowable=exc;
-	  };
-	};
-      };
-      cop=ncop;
-    };
-    return didsome;
+  /**
+   * Performs one optimization pass on the given list of operations.
+   * <P> Currently only evaluation of constant subexpressions is performed.
+   * @param code is a linked list of operations to optimize.
+   * @return true if additional optimization pass may simplify code further.
+   */
+  protected static boolean optimizeIteration(OPlist code) {
+    return code.optimize();
   };
 
   /**
@@ -422,17 +577,22 @@ public class Optimizer {
    * with the function, represented by this optimizer compiled in.
    */
   public CompiledExpression compile() {
+    return compileBits().getExpression();
+  };
+
+  /**
+   * Compiles the expression into an ExpressionBits object.
+   * <P>ExpressionBits is a wrapper allowing to store expression
+   * in any <TT>java.io.OutputStream</TT> using Java serialization mechanism.
+   * @return instance of the <TT>gnu.jel.ExpressionBits</TT> object.
+   * @see gnu.jel.ExpressionBits
+   */
+  public ExpressionBits compileBits() {
     seeFinished();
     ExpressionImage image=new ExpressionImage();
     code.compile(image);
-    //    if (caughtThrowable==null) 
     image.asm_return();
-      //    else {
-      //      image.asm_load_object(caughtThrowable);
-      //      image.asm_ThrowReturn();
-      //    };
-    ExpressionImage.dumpImage(image);
-    return image.getExpression();
+    return image.getBits();
   };
 
   /**
@@ -440,9 +600,7 @@ public class Optimizer {
    * @return String representation of the expression.
    */
   public String toString() {
-    //    if (caughtThrowable==null) 
     return code.toString();
-    //    return code.toString()+" $"+caughtThrowable.getClass().toString();
   };
 
   private void seeFinished() {
@@ -496,7 +654,7 @@ public class Optimizer {
 	op.load(2.0);
 	op.binaryOP_param();
 	op.load(2L);
-	op.binaryOP(ExpressionImage.BI_MU);
+	op.binaryOP(ExpressionImage.BI_MU,false);
 	op.finish();
 	Debug.println("");
 	Debug.println(op.toString());
@@ -518,14 +676,14 @@ public class Optimizer {
 	op.load(2.0);
 	op.binaryOP_param();
 	op.load(2L);
-	op.binaryOP(ExpressionImage.BI_MU);
+	op.binaryOP(ExpressionImage.BI_MU,false);
 	op.binaryOP_param();
 	   op.function_start();
 	   op.load(0);
 	   op.function_param();
            op.function_call("exp");
-	op.binaryOP(ExpressionImage.BI_MU);
-	op.binaryOP(ExpressionImage.BI_PL);
+	op.binaryOP(ExpressionImage.BI_MU,false);
+	op.binaryOP(ExpressionImage.BI_PL,false);
 	op.finish();
 	Debug.println(op.toString());
 	CompiledExpression ce=op.compile();
@@ -561,7 +719,9 @@ abstract class OP {
   protected OP prev=null;
   
   abstract void compile(ExpressionImage ei);
-
+  
+  boolean optimize(OPlist list) throws Throwable {return false;};
+  
   public OP next() { return next;};
 };
 
@@ -616,6 +776,18 @@ class OPlist extends OP {
     o.next=null; o.prev=null; // help GC
     size--;
   };
+
+  OPlist cut_end(OP from) {
+    OP curr=from;
+    OPlist newlist=new OPlist();
+    while (curr!=null) {
+      OP ncurr=curr.next;
+      remove(curr);
+      newlist.addLast(curr);
+      curr=ncurr;
+    };
+    return newlist;
+  };
   
   // This compiles whole expression
   void compile(ExpressionImage ei) {
@@ -626,6 +798,22 @@ class OPlist extends OP {
     };
   };
 
+  boolean optimize() {
+    try {
+      OP curr=next;
+      boolean didsome=false;
+      
+      while(curr!=null) {
+	OP ncurr=curr.next;
+	didsome = curr.optimize(this) || didsome;
+	curr=ncurr;
+      };
+      return didsome;
+    } catch (Throwable exc) {
+      return false;
+    };
+  };
+
   public String toString() {
     OP curr=next;
     StringBuffer res=new StringBuffer();
@@ -639,7 +827,7 @@ class OPlist extends OP {
   
 };
 
-class OP_start extends OP {  // Analogous to the list delimiter character
+class OP_start extends OP {  // Analogous to opening bracket in Lisp
   protected OP_function f=null;
   protected Method m=null;
   protected int id=999;
@@ -656,6 +844,8 @@ class OP_start extends OP {  // Analogous to the list delimiter character
   void compile(ExpressionImage ei) {
     if (m !=null) 
       ei.asm_func_start(m,id);
+    if (f instanceof OP_logical_not)
+      ei.asm_logical_block();
   };
 
   public String toString() {
@@ -667,9 +857,10 @@ class OP_start extends OP {  // Analogous to the list delimiter character
 // This class is a placeholder to insert the reference to the object, imple
 // memnting the function. As function overloading is allowed the actual
 // reference is unknown before ALL formal arguments are parsed.
-class OP_param extends OP_start {  // Analogous to opening bracket in Lisp
-  OP_function f=null;
+class OP_param extends OP_start { // Analogous to the list delimiter character
+  //  OP_function f=null;
   Class typeInStack=null;
+  boolean setbinaryparam=true;
   
   // f is the reference to the OP_function, which is the "closing bracket".
   OP_param(OP_function f,Class typeInStack) {
@@ -684,16 +875,30 @@ class OP_param extends OP_start {  // Analogous to opening bracket in Lisp
       } else {
 	list.addBefore(this,new OP_convert(cls)); // Add conversion OP
       };
+      typeInStack=cls;
     };
   };
-
+  
   void compile(ExpressionImage ei) {
     if (m!=null)
       ei.asm_func_param();
+    else {
+      if (setbinaryparam) {
+	if (f instanceof OP_binary) {
+	  if (((OP_binary)f).logical)
+	    ei.asm_logical_binary_param(((OP_binary)f).opc);
+	  else 
+	    ei.asm_binary_param(((OP_binary)f).opc);
+	};
+      };
+    };
   };
 
   public String toString() {
-    return ",";
+    if (setbinaryparam)
+      return ",";
+    else
+      return ".";
   };
  
 };
@@ -724,10 +929,12 @@ class OP_load extends OP {
 
   public String toString() {
 
-    if (type.isPrimitive()) {
+    if (type.isPrimitive())
       return what.toString()+typecodes[ExpressionImage.primitiveID(type)];
-    };
-
+    
+    if (what instanceof String)
+      return "\""+what+"\"";
+    
     return what.toString();
   };
 
@@ -735,6 +942,12 @@ class OP_load extends OP {
 };
 
 abstract class OP_function extends OP {
+
+  public boolean optimize(OPlist list) throws Throwable {
+    boolean can_interpret=canInterpret();
+    if (can_interpret) interpret(list);
+    return can_interpret;
+  };
   
   abstract boolean canInterpret();
   
@@ -824,6 +1037,118 @@ abstract class OP_function extends OP {
 };
 
 
+class OP_conditional extends OP_function {
+  OPlist trueList;
+  OPlist falseList;
+  
+  OP_conditional() {};
+  
+  public void setTrueList(OPlist trueList) {
+    this.trueList=trueList;
+  };
+  
+  public void setFalseList(OPlist falseList) {
+    this.falseList=falseList;
+  };
+
+  // should be called after both lists are set
+  // Ensures that types of both branches are the same
+  public void setType(Class type) {
+    
+    OP tll=trueList.getLast();
+    if (tll instanceof OP_convert) 
+      ((OP_convert)tll).setType(type);
+    else {
+      trueList.addLast(new OP_convert(type));
+    };
+
+    OP fll=falseList.getLast();
+    if (fll instanceof OP_convert) 
+      ((OP_convert)fll).setType(type);
+    else {
+      falseList.addLast(new OP_convert(type));
+    };
+    
+  };
+  
+  boolean canInterpret() {
+    return false;
+  };
+  
+  void interpret(OPlist list) throws Throwable {
+  };
+  
+  void compile(ExpressionImage ei) {
+    ei.asm_branch_start_true();
+    trueList.compile(ei);
+    ei.asm_branch_start_false();
+    falseList.compile(ei);
+    ei.asm_branch_end();
+  };
+
+  public boolean optimize(OPlist list) throws Throwable {
+    boolean optimized;
+    if (prev instanceof OP_load) {
+      boolean val=((Boolean)((OP_load)prev).what).booleanValue();
+      list.remove(prev); // get rid of used value
+
+      optimized=true;
+      OPlist integratelist;
+      if (val)
+	integratelist=trueList;
+      else
+	integratelist=falseList;
+      
+      integratelist.optimize();
+      OP iop=integratelist.getFirst();
+      // Integrate result into the main code
+      while (iop!=null) {
+	OP niop=iop.next;
+	integratelist.remove(iop);
+	list.addBefore(this,iop);
+	iop=niop;
+      };
+      list.remove(this);
+    } else {
+      optimized=trueList.optimize();
+      optimized=falseList.optimize() || optimized;
+    };
+    return optimized;
+  };
+
+  public String toString() {
+    return "[?"+trueList.toString()+":"+falseList.toString()+"]";
+  };
+};
+
+class OP_logical_not extends OP_function {
+  boolean canInterpret() {
+    return (prev instanceof OP_load) && (prev.prev instanceof OP_start);
+  };
+
+  void interpret(OPlist list) throws Throwable {
+    OP_load l=(OP_load) prev;
+    
+    // perform the complement
+    if (((Boolean)l.what).booleanValue()) 
+      l.what=Boolean.FALSE;
+    else 
+      l.what=Boolean.TRUE;
+    
+    list.remove(prev.prev); // OP_start
+    list.remove(this);
+  };
+  
+  void compile(ExpressionImage ei) {
+    ei.asm_logical_unblock_not();
+  };
+
+  public String toString() {
+    return "!]";
+  };
+ 
+};
+
 abstract class OP_unary_primitive extends OP_function {
 
   boolean canInterpret() {
@@ -897,17 +1222,32 @@ class OP_unary extends OP_unary_primitive {
     
     int id=ExpressionImage.primitiveID(c.type);
     
+    // Perform unary numeric promotion according to JLS
+    Class resClass=ExpressionImage.getUnaryPromoted(c.type);
+    if (Debug.enabled)
+      Debug.assert(resClass!=null,"Wrong type of operand for unary"+
+		   " operation ("+opc+") can't perform unary numeric "+
+		   "promotion.");
+    int resid=ExpressionImage.primitiveID(resClass);
+    
     Number val=widen(c.what,id);
-    if (opc==0) {
+    switch(opc) {
+    case 0:
       if (isFloat(id)) 
 	val=new Double(-val.doubleValue());
       else
 	val=new Long(-val.longValue());
-    } else {
+      break;
+    case 1:
+      val=new Long(~val.longValue());
+      break;
+    default:
       if (Debug.enabled)
 	Debug.println("Wrong unary opcode.");
     };
-    c.what=narrow(val,id);
+
+    c.what=narrow(val,resid);
+    c.type=resClass;
   };
   
   void compile(ExpressionImage ei) {
@@ -915,105 +1255,161 @@ class OP_unary extends OP_unary_primitive {
   };
 
   public String toString() {
-    return "-";
+    return ExpressionImage.unarySymbols[opc];
   };
   
 };
 
 class OP_binary extends OP_function {
   int opc=999;
+  boolean logical=false;
   OP_binary() {
   };
   
-  void setOperation(OPlist list, int opc,Class type) {
+  void setOperation(OPlist list, int opc,boolean logical,Class type) {
+    this.logical=logical;
     this.opc=opc;
-    int cpar=2;
-    OP cop=prev;
-    while ((cop!=null) && (cpar>=0)) {
-      if (cop instanceof OP_param) {
-	OP_param copp=(OP_param)cop;
-	if (copp.getFunction()==this) {
-	  copp.setConvert(list,type);
+    if (type!=Optimizer.string_class) { // For strings no conversions is needed
+      int cpar=2;
+      OP cop=prev;
+      while ((cop!=null) && (cpar>=0)) {
+	if (cop instanceof OP_param) {
+	  OP_param copp=(OP_param)cop;
+	  if (copp.getFunction()==this) {
+	    copp.setConvert(list,type);
+	  };
 	};
+	cop=cop.prev;
       };
-      cop=cop.prev;
     };
   };
 
   boolean canInterpret() { return operandsReady(2,false); };
   
   void interpret(OPlist list) throws Throwable {
+    if (logical) {
+      OP_load c2=(OP_load) prev.prev;
+      OP_load c1=(OP_load) prev.prev.prev.prev;
+      boolean b1=((Boolean)c1.what).booleanValue();
+      boolean b2=((Boolean)c2.what).booleanValue();
+      
+      switch(opc) {
+      case 0 : b1=b1&b2; break;
+      case 1 : b1=b1|b2; break;
+      default:
+	if (Debug.enabled)
+	  Debug.println("Wrong logical bitwise operation ("+opc+").");      
+      };
+     
+      if (b1) c1.what=Boolean.TRUE; else c1.what=Boolean.FALSE;
+      
+      consume(list,c1,2,false);      
+      return;
+    };
+    
     OP_load c2=(OP_load) prev.prev;
     OP_load c1=(OP_load) prev.prev.prev.prev;
     
-    if (Debug.enabled)
-      Debug.assert((c1.type!=null) && (c1.type==c2.type) && 
-		   (c2.type.isPrimitive()),"Wrong types for binary OP");
-    
-    int id=ExpressionImage.primitiveID(c1.type);
-    
-    // Widen
-    Number n1=widen(c1.what,id),n2=widen(c2.what,id);
-
-    // Perform      
-    if (isFloat(id)) {
-      double d1=n1.doubleValue(),d2=n2.doubleValue();
-      switch (opc) {
-      case 0: d1=d1+d2; break; //PL
-      case 1: d1=d1-d2; break; //MI
-      case 2: d1=d1*d2; break; //MU
-      case 3: d1=d1/d2; break; //DI
-      case 4: d1=d1%d2; break; //RE
-      default :
-	if (Debug.enabled)
+    if ((opc==ExpressionImage.BI_PL) && (c1.type==Optimizer.string_class)) {
+      // String concatenation
+      c1.what=((String)c1.what)+String.valueOf(c2.what);
+    } else {
+      // Operators on primitive types
+      if (Debug.enabled)
+	Debug.assert((c1.type!=null) && (c1.type==c2.type) && 
+		     (c2.type.isPrimitive()),"Wrong types for binary OP");
+      
+      int id=ExpressionImage.primitiveID(c1.type);
+      
+      // Widen
+      Number n1=widen(c1.what,id),n2=widen(c2.what,id);
+      
+      // Perform      
+      boolean boolres=false;
+      boolean resbool=false;
+      if (isFloat(id)) {
+	double d1=n1.doubleValue(),d2=n2.doubleValue();
+	boolean wrop=false;
+	switch (opc) {
+	case 0 : d1=d1+d2; break; //PL
+	case 1 : d1=d1-d2; break; //MI
+	case 2 : d1=d1*d2; break; //MU
+	case 3 : d1=d1/d2; break; //DI
+	case 4 : d1=d1%d2; break; //RE
+	case 5 : wrop=true;break; //AN
+	case 6 : wrop=true;break; //OR
+	case 7 : wrop=true;break; //XO
+	case 8 : boolres=true; resbool=(d1==d2); break; //EQ
+	case 9 : boolres=true; resbool=(d1!=d2); break; //NE
+	case 10: boolres=true; resbool=(d1<d2);  break; //LT
+	case 11: boolres=true; resbool=(d1>=d2); break; //GE
+	case 12: boolres=true; resbool=(d1>d2);  break; //GT
+	case 13: boolres=true; resbool=(d1<=d2); break; //LE
+	default :
+	  wrop=true;
+	};
+	if (Debug.enabled && wrop)
 	  Debug.println("Wrong operation on float ("+opc+").");      
+	
+	if (!boolres) n1=new Double(d1);
+	else { // booleans are represented by longs temporarily
+	  if (resbool) n1=new Long(1L); else n1=new Long(0);
+	};
+      } else { 
+	long l1=n1.longValue(),l2=n2.longValue();
+	switch (opc) {
+	case 0: l1=l1+l2;break; //PL
+	case 1: l1=l1-l2;break; //MI
+	case 2: l1=l1*l2; break; //MU
+	case 3: l1=l1/l2; break; //DI
+	case 4: l1=l1%l2; break; //RE
+	case 5: l1=l1&l2; break; //AN
+	case 6: l1=l1|l2; break; //OR
+	case 7: l1=l1^l2; break; //XO
+	case 8 : boolres=true; l1=(l1==l2?1L:0L); break; //EQ
+	case 9 : boolres=true; l1=(l1!=l2?1L:0L); break; //NE
+	case 10: boolres=true; l1=(l1< l2?1L:0L); break; //LT
+	case 11: boolres=true; l1=(l1>=l2?1L:0L); break; //GE
+	case 12: boolres=true; l1=(l1> l2?1L:0L); break; //GT
+	case 13: boolres=true; l1=(l1<=l2?1L:0L); break; //LE
+	case 14: l1=l1<<l2; break;  // LS
+	case 15: l1=l1>>l2; break;  // RS
+	case 16: { // for this kind of shifts the bit width of variable is
+	  // important
+	  if (id==4) // == because there is unary numeric promotion before op
+	    l1=(int)l1>>>l2; 
+	  else 
+	    l1=l1>>>l2; 
+	  break; 
+	}; // RUS
+	default :
+	  if (Debug.enabled)
+	    Debug.println("Wrong operation on integer ("+opc+").");      
+	};
+	n1=new Long(l1);
       };
-      n1=new Double(d1);
-    } else { 
-      long l1=n1.longValue(),l2=n2.longValue();
-      switch (opc) {
-      case 0: l1=l1+l2;break; //PL
-      case 1: l1=l1-l2;break; //MI
-      case 2: l1=l1*l2; break; //MU
-      case 3: l1=l1/l2; break; //DI
-      case 4: l1=l1%l2; break; //RE
-      case 5: l1=l1&l2; break; //AN
-      case 6: l1=l1|l2; break; //OR
-      case 7: l1=l1^l2; break; //XO
-      default :
-	if (Debug.enabled)
-	  Debug.println("Wrong operation on integer ("+opc+").");      
-      };
-      n1=new Long(l1);
+      
+      // Narrow    
+      if (boolres) {
+	c1.what=narrow(n1,0);
+	c1.type=Boolean.TYPE;
+      } else c1.what=narrow(n1,id);
     };
-
-    // Narrow    
-    c1.what=narrow(n1,id);
-    
     consume(list,c1,2,false);
   };
 
   void compile(ExpressionImage ei) {
-    ei.asm_binary(opc);
+    if (logical)
+      ei.asm_logical_binary(opc);
+    else
+      ei.asm_binary(opc);
   };
 
   public String toString() {
-    char ch='?';
-    switch (opc) {
-    case 0: ch='+';break; //PL
-    case 1: ch='-';break; //MI
-    case 2: ch='*'; break; //MU
-    case 3: ch='/'; break; //DI
-    case 4: ch='%'; break; //RE
-    case 5: ch='&'; break; //AN
-    case 6: ch='|'; break; //OR
-    case 7: ch='^'; break; //XO
-    default :
-      if (Debug.enabled)
-	Debug.println("Wrong operation on integer ("+opc+").");      
-    };
-    
-    return ""+ch;
+    if (logical)
+      return ExpressionImage.logicalSymbols[opc];
+    else
+      return ExpressionImage.binarySymbols[opc];
   };
 
   
@@ -1096,7 +1492,6 @@ class OP_call extends OP_function {
   public String toString() {
     return m.getName()+"]";
   };
-  
 }
 
 
